from __future__ import unicode_literals, print_function, absolute_import
from .. import JiraCommand


class Comment(JiraCommand):
    """
    Comment an issue to someone
    """

    def add_arguments(self):
        self.parser.add_argument(
            'body',
            help='The comment content.'
        )

        self.parser.add_argument(
            '-k',
            '--key',
            help='The issue key to use. Defaults to the name of the active branch.'
        )

    def run(self, args):
        key = args.key or self.giles.repo.active_branch.name
        key = key.upper()
        issue = self.giles.jira.issue(key)
        self.giles.jira.add_comment(issue, args.body)
        return {
            'message': 'Success: Added comment on {} '.format(key)
        }



from __future__ import unicode_literals, print_function, absolute_import
from .. import GilesCommand, RC, PROFILE

from giles.utils import get_input
import chalk
try:
    import ConfigParser as configparser
except ImportError:  # pragma: no cover
    import configparser
import os
import stat
import textwrap


class Setup(GilesCommand):
    """Configure your Giles workflow manager."""

    def add_arguments(self):
        self.parser.add_argument(
            '-d',
            '--dry-run',
            help='Setup both the .gilesrc and .gilesrc_profile',
            action='store_true'
        )
        self.parser.add_argument(
            '-r',
            '--rc',
            help='Setup only the .gilesrc details.',
            action='store_true'
        )
        self.parser.add_argument(
            '-p',
            '--profile',
            help='Setup only the.gilesrc_profile details.',
            action='store_true'
        )

    def run(self, args):
        welcome_msg = """
        ########################################################
        #
        # Giles configuration
        #
        ########################################################"""

        chalk.green(textwrap.dedent(welcome_msg), opts=('bold',))
        try:
            self._run_configs(args)
        except KeyboardInterrupt:
            raise KeyboardInterrupt('\n\n!!Error: Configuration Incomplete!!')

    def _file_check(self, name):
        exists = os.path.isfile(name)
        if exists:
            chalk.yellow(
                '\nWARNING: "{}" already exists.'.format(name),
                opts=['bold']
            )
            _override = None
            while _override not in ('y', 'n'):
                _input = get_input(
                    'Override "{}" [y/N]: '.format(name)
                )
                _override = _input or 'n'
            if not _override or _override != 'y':
                raise StopIteration()
        return exists

    def _rc_config(self, args, _all=False):
        if _all or args.rc:
            exists = self._file_check(RC)
            chalk.magenta(
                "\nPlease setup your .gilesrc details",
                opts=('bold',)
            )
            jira_url = get_input(
                'JIRA URL (e.g. https://example.atlassian.net): ',
                required=True
            )
            github_project = get_input(
                'GitHub Project (e.g. anthonyalmarza/giles): ',
                required=True
            )
            github_repo = get_input('Git Repo Path (default: "."): ')
            github_branch = get_input(
                'Git Default Branch (master): '
            )
            if not args.dry_run:
                config = configparser.RawConfigParser()
                config.add_section('jira')
                config.add_section('github')
                config.set('jira', 'url', jira_url)
                config.set('github', 'project', github_project)
                config.set('github', 'repo_path', github_repo or '.')
                config.set('github', 'branch', github_branch or 'master')
                # Writing our configuration file to 'example.cfg'
                with open(RC, 'wb') as configfile:
                    config.write(configfile)
            else:
                chalk.blue('DryRun:', opts=['bold'])

            chalk.blue(
                '"{}" setup complete'.format(RC),
                opts=['bold']
            )

    def _profile_config(self, args, _all=False):
        if _all or args.profile:
            exists = self._file_check(PROFILE)
            chalk.magenta(
                "\nPlease setup your .gilesrc_profile details",
                opts=('bold',)
            )
            jira_username = get_input(
                'JIRA Username: ',
                required=True
            )
            jira_password = get_input(
                'JIRA Password: ',
                required=True,
                secret=True
            )

            github_username = get_input(
                'GitHub Username: ',
                required=True
            )
            github_password = get_input(
                'GitHub Password: ',
                required=True,
                secret=True
            )
            if not args.dry_run:
                profile = configparser.RawConfigParser()
                profile.add_section('jira')
                profile.add_section('github')
                profile.set('jira', 'username', jira_username)
                profile.set('jira', 'password', jira_password)
                profile.set('github', 'username', github_username)
                profile.set('github', 'password', github_password)
                if exists:
                    os.remove(PROFILE)
                with open(PROFILE, 'wb') as configfile:
                    profile.write(configfile)
                    os.chmod(PROFILE, stat.S_IREAD)
            else:
                chalk.blue('DryRun:', opts=['bold'])
            chalk.blue(
                '"{}" setup complete'.format(PROFILE),
                opts=['bold']
            )

    def _run_configs(self, args):
        _all = not args.rc and not args.profile
        for config in (self._rc_config, self._profile_config):
            try:
                config(args, _all=_all)
            except StopIteration:
                pass


from __future__ import unicode_literals, print_function, absolute_import
from .. import JiraCommand


class Claim(JiraCommand):
    """
    Claim a JIRA ticket
    """
    def add_parsers(self):
        self.parser.add_argument(
            '-k',
            '--key',
            help='The issue key to use. Defaults to the name of the active branch.'
        )

        self.parser.add_argument(
            '-m',
            '--message',
            help='Add an optional comment to add to the ticket.'
        )

    def run(self, args):
        key = args.key or self.giles.repo.active_branch.name
        key = key.upper()
        issue = self.giles.jira.issue(key)
        self.giles.jira.assign_issue(issue, self.giles.jira.current_user())
        if args.message:
            self.giles.jira.add_comment(issue, args.message)
        return {
            'message': 'Success: Claimed {} '.format(key)
        }



from __future__ import unicode_literals, print_function, absolute_import
# http://gitpython.readthedocs.io/en/stable/
# http://jira.readthedocs.io/en/latest
# http://pygithub.readthedocs.io/en/latest/introduction.html
try:
    import ConfigParser as configparser
except ImportError:  # pragma: no cover
    import configparser

from giles.commands.bases import CommandBase, SubCommandBase
from github import Github
from jira import JIRA
import git
import os

RC = os.environ.get('GILESRC', '.gilesrc')
PROFILE = os.environ.get('GILESRCPROFILE', '.gilesrc_profile')


class Giles(CommandBase):
    """
        Giles is a terminal based workflow manager that integrates JIRA with
        both slack and github.
    """
    _is_configured = False
    _is_setup = False

    def __init__(self, configure=False):
        if configure:
            self._configure()

    def _configure(self):
        if not self._is_configured:
            self.config = configparser.SafeConfigParser()
            self.profile = configparser.SafeConfigParser()

            if os.path.isfile(RC) and os.path.isfile(PROFILE):
                self._is_setup = True

            if self._is_setup:
                self.config.read(RC)
                self.profile.read(PROFILE)

                self.default_branch = self.config.get('github', 'branch')
                self.project = self.config.get('github', 'project')
                self.repo = git.Repo(self.config.get('github', 'repo_path'))
                self.git = self.repo.git
                self._is_configured = True

    @property
    def github(self):
        if not hasattr(self, '_github'):
            self._github = Github(
                self.profile.get('github', 'username'),
                self.profile.get('github', 'password')
            )
        return self._github

    @property
    def jira(self):
        if not hasattr(self, '_jira'):
            self._jira = JIRA(
                self.config.get('jira', 'url'),
                basic_auth=(
                    self.profile.get('jira', 'username'),
                    self.profile.get('jira', 'password')
                )
            )
        return self._jira

class GilesCommand(SubCommandBase):
    abstract = True
    parent = Giles



from __future__ import unicode_literals, print_function, absolute_import
from giles.commands.bases import CommandBase, SubCommandBase
from giles.commands.giles import Giles
# http://jira.readthedocs.io/en/latest/api.html#jira.JIRA.search_assignable_users_for_issues


class Jira(CommandBase):
    giles = Giles(configure=True)


class JiraCommand(SubCommandBase):

    parent = Jira
    abstract = True

    @property
    def giles(self):
        return self.parent.giles


from __future__ import unicode_literals, print_function, absolute_import
from .. import GitHubCommand


class Submit(GitHubCommand):

    def add_arguments(self):
        self.parser.add_argument(
            'title',
            help='The title to use for the pull request.'
        )

        self.parser.add_argument(
            '--head',
            default=None,
            help=(
                'Name of the branch, commit or tag that you want to use in the PR. '
                'Note that this is the "from" branch e.g. PR from MYBRANCH to master'
            )
        )

        self.parser.add_argument(
            '-b',
            '--base',
            help=(
                'Base is the branch you are submitting the PR to. Defaults to the '
                'name of the default branch.'
            )
        )


        self.parser.add_argument(
            '-m',
            '--message',
            help='Optional message to include in your PR.'
        )

    def run(self, args):
        head = args.head or self.giles.repo.active_branch.name
        base = args.base or self.giles.default_branch
        repo = self.giles.github.get_repo(self.giles.project)
        response = repo.create_pull(
            title=args.title,
            body=args.message or '',
            base=base,
            head=head,
        )
        return {
            'message': 'Success: Created PR#{} from "{}" to "{}"'.format(
                response.number, head, base
            )
        }



from __future__ import unicode_literals, print_function, absolute_import
from giles.utils import load_subcommands


for name in ('giles', 'jira', 'github'):
    load_subcommands('giles.commands.' + name)


from __future__ import unicode_literals, print_function, absolute_import
import chalk
import getpass
import importlib
import os
import sys


def get_input(prompt, required=False, secret=False):
    output = None
    _input = getpass.getpass if secret else raw_input

    formatted_prompt = chalk.format_txt('white', prompt, None, ['bold'])
    output = _input(formatted_prompt)
    while not output and required:
        output = _input(formatted_prompt)
    return output


def load_subcommands(dot_path):
    importlib.import_module(dot_path)
    module = sys.modules[dot_path]
    path = module.__path__[0]
    for name in os.listdir(os.path.join(path, 'subcommands')):
        if not name.startswith('_') and name.endswith('py'):
            importlib.import_module(
                dot_path + '.subcommands.' + name.replace('.py', '')
            )



from __future__ import unicode_literals, print_function, absolute_import
from .. import JiraCommand


class List(JiraCommand):

    def add_arguments(self):
        self.parser.add_argument(
            '--users',
            help='List all of the users',
            action='store_true'
        )

        self.parser.add_argument(
            '--comments',
            help=(
                'List all of the comments for the given issue key.'
                'Defaults to the current active branch name.'
            ),
            default='active_branch'
        )

        self.parser.add_argument(
            '--search',
            help='List issues using the provided JQL string'
        )

        self.parser.add_argument(
            '--filter',
            help='List issues using the provided filter id'
        )

        self.parser.add_argument(
            '--favorite-filters',
            help='List all of your favorite filters by id and name.',
            action='store_true'
        )

        self.parser.add_argument(
            '--page',
            help='',
            default=0,
            type=int
        )

        self.parser.add_argument(
            '--limit',
            help='',
            type=int
        )

    def run(self, args):
        opts = {'page': args.page}
        if args.limit:
            opts['limit'] = args.limit

        if args.users:
            table = {
                'heading': ['NAME', 'KEY', 'ACTIVE'],
                'data': [
                    (user.displayName, user.key, user.active)
                    for user in self.users(**opts)
                ],
                'width': 24
            }
        elif args.search:
            table = {
                'heading': ['KEY', 'SUMMARY'],
                'data': [
                    (issue.key, issue.fields.summary)
                    for issue in self.search(args.search, **opts)
                ]
            }
        elif args.favorite_filters:
            table = {
                'heading': ['ID', 'NAME'],
                'data': [
                    (_filter.id, _filter.name)
                    for _filter in self.favorite_filters()
                ],
                'width': 5
            }
        elif args.filter:
            _filter = self.giles.jira.filter(args.filter)
            table = {
                'heading': ['KEY', 'SUMMARY'],
                'data': [
                    (issue.key, issue.fields.summary)
                    for issue in self.search(_filter.jql, **opts)
                ]
            }
        elif args.comments:
            branch = self.giles.repo.active_branch.name
            key = args.comments if args.comments != 'active_branch' else branch
            issue = self.giles.jira.issue(key)
            comments = self.giles.jira.comments(issue)
            table = {
                'heading': ['ID', 'AUTHOR', 'BODY'],
                'data': [
                    (comment.id, comment.author.key, comment.body)
                    for comment in comments
                ]
            }

        else:
            raise RuntimeError('Not a doctor...')

        return {'table': table}

    def search(self, jql, page=0, limit=5):
        offset = page * limit
        return self.giles.jira.search_issues(
            jql,
            startAt=offset,
            maxResults=limit
        )

    def users(self, page=0, limit=50):
        offset = page * limit
        return self.giles.jira.search_users(
            '_',
            startAt=offset,
            maxResults=limit
        )

    def favorite_filters(self):
        return self.giles.jira.favourite_filters()



from __future__ import unicode_literals, print_function, absolute_import
from giles.commands.bases import CommandBase, SubCommandBase
from giles.commands.giles import Giles


class GitHub(CommandBase):
    giles = Giles(configure=True)


class GitHubCommand(SubCommandBase):
    parent = GitHub
    abstract = True

    @property
    def giles(self):
        return self.parent.giles


from __future__ import unicode_literals, print_function, absolute_import
from .. import JiraCommand


class Assign(JiraCommand):
    """
    Assign an issue to someone
    """

    def add_arguments(self):
        self.parser.add_argument(
            'name',
            help='The name of the assignee to use.'
        )

        self.parser.add_argument(
            '-k',
            '--key',
            help='The issue key to use. Defaults to the name of the active branch.'
        )
        self.parser.add_argument(
            '-m',
            '--message',
            help='Add a comment'
        )


    def run(self, args):
        key = args.key or self.giles.repo.active_branch.name
        key = key.upper()
        issue = self.giles.jira.issue(key)
        self.giles.jira.assign_issue(issue, args.name)
        if args.message:
            self.giles.jira.add_comment(issue, args.message)
        return {
            'message': 'Success: Assigned "{}" on {} '.format(args.name, key)
        }



from __future__ import unicode_literals, print_function, absolute_import
import argparse
import chalk
import six

# https://docs.python.org/2/library/argparse.html

class CommandTypeBase(type):

    def __new__(cls, name, bases, attrs):
        is_abstract = attrs.pop('abstract', False)
        new_class = super(CommandTypeBase, cls).__new__(cls, name, bases, attrs)
        parents = [base for base in bases if isinstance(base, CommandTypeBase)]
        if parents:
            new_class._instance = None
            new_class.is_abstract = is_abstract
            cls.setup(new_class)
        return new_class

    def setup(cls):
        raise NotImplementedError()


class CommandType(CommandTypeBase):

    def setup(cls):
        parser = argparse.ArgumentParser(
            prog=cls.__name__.lower(),
            description=cls.__doc__
        )

        cls.subparses = parser.add_subparsers(help='Summary')
        cls.parser = parser


class SubCommandType(CommandTypeBase):

    def setup(cls):
        if not cls.is_abstract:
            cls.parser = cls.parent.subparses.add_parser(
                cls.__name__.lower(),
                description=cls.__doc__
            )
            # prime the singleton
            cls()


class CommandBase(six.with_metaclass(CommandType)):

    additional_commands = []

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def run(self):
        args = self.parser.parse_args()
        args.func(args)


class SubCommandBase(six.with_metaclass(SubCommandType)):
    """
    """

    _instance = None
    parent = NotImplemented

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        self.parser.set_defaults(func=self._run)
        self.add_arguments()

    def add_arguments(self):
        pass

    def run(self, args):
        raise NotImplementedError()

    def render(self, output):
        output = output or {}
        _chalk = getattr(chalk, output.get('color', 'blue'))
        if 'message' in output:
            _chalk('{}'.format(output.get('message')))

        table = output.get('table')
        if table:
            width = table.get('width', 12)
            heading = self.format_row(table.get('heading', []), width=width)
            _chalk(heading, opts=['bold'])
            for item in table.get('data', []):
                _chalk(self.format_row(item, width=width))

    def format_row(self, row, width):
        return '\t'.join([str(_).ljust(width) for _ in row])

    def _run(self, args):
        try:
            output = self.run(args)
            self.render(output)
        except Exception as err:
            chalk.red(str(err), opts=['bold'])
            raise SystemError(err)



from __future__ import unicode_literals, print_function, absolute_import
from setuptools import setup, find_packages
import giles


setup(
    author="Anthony Almarza",
    name="giles",
    version=giles.__version__,
    packages=find_packages(exclude=["test*", ]),
    url="https://github.com/anthonyalmarza/giles",
    description=(
        "``giles`` "
    ),
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    keywords=['giles', 'jira', 'workflow', ],
    install_requires=[
        'jira',
        'GitPython',
        'PyGithub',
        'pychalk',
        'slackclient',
        'urllib3[secure]'
    ],
    extras_require={'dev': ['ipdb', 'mock', 'tox', 'coverage']},
    include_package_data=True,
    scripts=[
        'bin/giles',
        'bin/jira',
        'bin/github',
    ]
)



from __future__ import unicode_literals, print_function, absolute_import
from .. import GitHubCommand


class Branch(GitHubCommand):

    def add_arguments(self):
        self.parser.add_argument(
            'name',
            help='The name of the branch to use.'
        )

        self.parser.add_argument(
            '-c',
            '--checkout',
            help='The branch to checkout. Defaults to the name of the default branch.'
        )

    def run(self, args):
        checkout_branch = args.checkout or self.giles.default_branch
        self.giles.git.checkout(checkout_branch)
        self.giles.git.fetch()
        self.giles.git.checkout('-b', args.name)

        return {
            'message': 'Success: Created branch "{}" from "{}"'.format(
                args.name, checkout_branch
            )
        }
